/*
 * freescope - Free source browser
 * Copyright (C) 2001  Olivier Deme
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

// FILE: fconfig.cpp

/************/
/* INCLUDES */
/************/

#include "fconfig.h"
#include "readconfig.h"
#include "defs.h"


/**************************/
/* STATIC INITIALIZATIONS */
/**************************/

FConfig* FConfig::m_instance = NULL;

/************************/
/* FUNCTION DEFINITIONS */
/************************/

/*
 * FUNCTION:    FConfig::FConfig
 *
 * DESCRIPTION: Constructor. Initialise all configuration attributes
 *              with default values
 *
 * IN:
 * IN-OUT:
 * RETURN CODE:
 */

FConfig::FConfig ()
{
    m_db_file                  = "freescope.out";
    m_input_file               = "freescope.files";
    m_update_on_startup        = 1;
    m_key_next_selection       = 'k';
    m_key_prev_selection       = 'j';
    m_key_next_query           = '\t';
    m_key_prev_query           = '\0';
    m_key_next_page            = 'l';
    m_key_prev_page            = 'h';
    m_key_switch_mode          = ctrl('w');
    m_key_next_completion      = ctrl('n');
    m_key_prev_completion      = ctrl('p');
    m_key_rebuild_db           = ctrl('r');
    m_key_help                 = '?';
    m_key_escape               = '\\';
    m_key_history_forward      = ctrl('f');
    m_key_history_backward     = ctrl('b');
    g_set_db_file              = FConfig::set_db_file;
    g_set_input_file           = FConfig::set_input_file;
    g_set_update_on_startup    = FConfig::set_update_on_startup;
    g_set_key_next_selection   = FConfig::set_key_next_selection;
    g_set_key_prev_selection   = FConfig::set_key_prev_selection;
    g_set_key_next_page        = FConfig::set_key_next_page;
    g_set_key_prev_page        = FConfig::set_key_prev_page;
    g_set_key_next_query       = FConfig::set_key_next_query;
    g_set_key_prev_query       = FConfig::set_key_prev_query;
    g_set_key_switch_mode      = FConfig::set_key_switch_mode;
    g_set_key_next_completion  = FConfig::set_key_next_completion;
    g_set_key_prev_completion  = FConfig::set_key_prev_completion;
    g_set_key_rebuild_db       = FConfig::set_key_rebuild_db;
    g_set_key_help             = FConfig::set_key_help;
    g_set_key_history_forward  = FConfig::set_key_history_forward;
    g_set_key_history_backward = FConfig::set_key_history_backward;
    m_dumping_db               = false;
}

/*
 * FUNCTION:    FConfig::instance
 *
 * DESCRIPTION: return a pointer to the instance.
 *
 * IN:  
 * IN-OUT:
 * RETURN CODE: A pointer to the instance
 */

FConfig* FConfig::instance ()
{
    if (m_instance == NULL)
    {
        m_instance = new FConfig();
        assert(m_instance != 0);
    }

    return m_instance;
}

/*
 * FUNCTION:    FConfig::init_config
 *
 * DESCRIPTION: This function reads the given configuration file.
 *
 * IN:          config_file     The configuration file to read
 * IN-OUT:
 * OUT:
 * RETURN CODE:
 * SIDE EFFECT:
 */

void FConfig::init_config (const string& config_file)
{
    read_config((char*)config_file.c_str());
}

/*
 * FUNCTION:    FConfig::set_db_file
 *
 * DESCRIPTION: This function sets the database file attribute
 *
 * IN:          db_file         The database file
 * IN-OUT:
 * OUT:
 * RETURN CODE:
 * SIDE EFFECT:
 */

void FConfig::set_db_file (char* db_file)
{
    m_instance->m_db_file = db_file;
}

/*
 * FUNCTION:    FConfig::get_db_file
 *
 * DESCRIPTION: This function gets the database file attribute
 *
 * IN:        
 * IN-OUT:
 * OUT:
 * RETURN CODE: The database file name
 * SIDE EFFECT:
 */

string FConfig::get_db_file ()
{
    return m_db_file;
}

/*
 * FUNCTION:    FConfig::set_input_file
 *
 * DESCRIPTION: This function sets the input file attribute
 *
 * IN:          input_file         The database file
 * IN-OUT:
 * OUT:
 * RETURN CODE:
 * SIDE EFFECT:
 */

void FConfig::set_input_file (char* input_file)
{
    m_instance->m_input_file = input_file;
}

/*
 * FUNCTION:    FConfig::get_input_file
 *
 * DESCRIPTION: This function gets the input file attribute
 *
 * IN:        
 * IN-OUT:
 * OUT:
 * RETURN CODE: The input file name
 * SIDE EFFECT:
 */

string FConfig::get_input_file ()
{
    return m_input_file;
}

/*
 * FUNCTION:    FConfig::get_input_file_str
 *
 * DESCRIPTION: This function gets the input file attribute by returning a
 *              char*
 *
 * IN:        
 * IN-OUT:
 * OUT:
 * RETURN CODE: The input file name
 * SIDE EFFECT:
 */

char* FConfig::get_input_file_str ()
{
    return (char*)m_input_file.c_str();
}


/*
 * FUNCTION:    FConfig::set_update_on_startup
 *
 * DESCRIPTION: This function sets the update on startup attribute
 *
 * IN:          update_on_startup         true or false
 * IN-OUT:
 * OUT:
 * RETURN CODE:
 * SIDE EFFECT:
 */

void FConfig::set_update_on_startup (int update_on_startup)
{
    m_instance->m_update_on_startup = update_on_startup;
}

/*
 * FUNCTION:    FConfig::get_update_on_startup
 *
 * DESCRIPTION: This function gets the update on startup attribute
 *
 * IN:        
 * IN-OUT:
 * OUT:
 * RETURN CODE: The database file name
 * SIDE EFFECT:
 */

int FConfig::get_update_on_startup ()
{
    return m_update_on_startup;
}

/*
 * FUNCTION:    FConfig::is_valid_key
 *
 * DESCRIPTION: This function verifies if the given key is valid and if
 *              it doesn't conflict with another key mapping.
 *
 * IN:          key             The key to check
 *              type            The type of mapping
 * IN-OUT:
 * OUT:
 * RETURN CODE: true if the key is valid.
 * SIDE EFFECT:
 */

bool FConfig::is_valid_key (char     key,
                            KEY_TYPE type)
{
    // CTRL-C, CTRL-D, CTRL-Z CANNOT BE REMAPPED!
    if (key == ctrl('c') || key == ctrl('d') || key == ctrl('z'))
        return false;
    
    // TODO: Add checking for possible mapping conflicts.

    return true;
}

/*
 * FUNCTION:    FConfig::set_key_next_selection
 *
 * DESCRIPTION: This function defines the key for slecting next match while
 *              browsing results.
 *
 * IN:          key             The key to map
 * IN-OUT:
 * OUT:
 * RETURN CODE: 0 if the key could be mapped.
 * SIDE EFFECT:
 */

int FConfig::set_key_next_selection (char key)
{
    // Let's check if the key causes conflict.
    if (m_instance->is_valid_key(key, KEY_NEXT_SELECTION) == false)
        return -1;
    
    m_instance->m_key_next_selection = key;

    return 0;
}

/*
 * FUNCTION:    FConfig::get_key_next_selection
 *
 * DESCRIPTION: This function returns the character corresponding to 
 *              the next selection key.
 *
 * IN: 
 * IN-OUT:
 * OUT:
 * RETURN CODE: The next selection key
 * SIDE EFFECT:
 */

char FConfig::get_key_next_selection ()
{
    return m_key_next_selection;
}

/*
 * FUNCTION:    FConfig::set_key_prev_selection
 *
 * DESCRIPTION: This function defines the key for slecting prev match while
 *              browsing results.
 *
 * IN:          key             The key to map
 * IN-OUT:
 * OUT:
 * RETURN CODE: 0 if the key could be mapped.
 * SIDE EFFECT:
 */

int FConfig::set_key_prev_selection (char key)
{
    // Let's check if the key causes conflict.
    if (m_instance->is_valid_key(key, KEY_PREV_SELECTION) == false)
        return -1;
    
    m_instance->m_key_prev_selection = key;

    return 0;
}

/*
 * FUNCTION:    FConfig::get_key_prev_selection
 *
 * DESCRIPTION: This function returns the character corresponding to 
 *              the prev selection key.
 *
 * IN: 
 * IN-OUT:
 * OUT:
 * RETURN CODE: The prev selection key
 * SIDE EFFECT:
 */

char FConfig::get_key_prev_selection ()
{
    return m_key_prev_selection;
}

/*
 * FUNCTION:    FConfig::set_key_next_page
 *
 * DESCRIPTION: This function defines the key for going to next page of 
 *              results.
 *
 * IN:          key             The key to map
 * IN-OUT:
 * OUT:
 * RETURN CODE: 0 if the key could be mapped.
 * SIDE EFFECT:
 */

int FConfig::set_key_next_page (char key)
{
    // Let's check if the key causes conflict.
    if (m_instance->is_valid_key(key, KEY_NEXT_PAGE) == false)
        return -1;
    
    m_instance->m_key_next_page = key;

    return 0;
}

/*
 * FUNCTION:    FConfig::get_key_next_page
 *
 * DESCRIPTION: This function returns the character corresponding to 
 *              the next result page key.
 *
 * IN: 
 * IN-OUT:
 * OUT:
 * RETURN CODE: The prev page key
 * SIDE EFFECT:
 */

char FConfig::get_key_next_page ()
{
    return m_key_next_page;
}

/*
 * FUNCTION:    FConfig::set_key_prev_page
 *
 * DESCRIPTION: This function defines the key for going to prev page of 
 *              results.
 *
 * IN:          key             The key to map
 * IN-OUT:
 * OUT:
 * RETURN CODE: 0 if the key could be mapped.
 * SIDE EFFECT:
 */

int FConfig::set_key_prev_page (char key)
{
    // Let's check if the key causes conflict.
    if (m_instance->is_valid_key(key, KEY_PREV_PAGE) == false)
        return -1;
    
    m_instance->m_key_prev_page = key;

    return 0;
}

/*
 * FUNCTION:    FConfig::get_key_prev_page
 *
 * DESCRIPTION: This function returns the character corresponding to 
 *              the prev result page key.
 *
 * IN: 
 * IN-OUT:
 * OUT:
 * RETURN CODE: The prev page key
 * SIDE EFFECT:
 */

char FConfig::get_key_prev_page ()
{
    return m_key_prev_page;
}

/*
 * FUNCTION:    FConfig::set_key_next_query
 *
 * DESCRIPTION: This function defines the key for selecting the next query
 *              type.
 *
 * IN:          key             The key to map
 * IN-OUT:
 * OUT:
 * RETURN CODE: 0 if the key could be mapped.
 * SIDE EFFECT:
 */

int FConfig::set_key_next_query (char key)
{
    // Let's check if the key causes conflict.
    if (m_instance->is_valid_key(key, KEY_NEXT_QUERY) == false)
        return -1;
    
    m_instance->m_key_next_query = key;

    return 0;
}

/*
 * FUNCTION:    FConfig::get_key_next_query
 *
 * DESCRIPTION: This function returns the character corresponding to 
 *              the next query key.
 *
 * IN: 
 * IN-OUT:
 * OUT:
 * RETURN CODE: The next query key
 * SIDE EFFECT:
 */

char FConfig::get_key_next_query ()
{
    return m_key_next_query;
}

/*
 * FUNCTION:    FConfig::set_key_prev_query
 *
 * DESCRIPTION: This function defines the key for selecting the prev query
 *              type.
 *
 * IN:          key             The key to map
 * IN-OUT:
 * OUT:
 * RETURN CODE: 0 if the key could be mapped.
 * SIDE EFFECT:
 */

int FConfig::set_key_prev_query (char key)
{
    // Let's check if the key causes conflict.
    if (m_instance->is_valid_key(key, KEY_PREV_QUERY) == false)
        return -1;
    
    m_instance->m_key_prev_query = key;

    return 0;
}

/*
 * FUNCTION:    FConfig::get_key_prev_query
 *
 * DESCRIPTION: This function returns the character corresponding to 
 *              the prev query key.
 *
 * IN: 
 * IN-OUT:
 * OUT:
 * RETURN CODE: The prev query key
 * SIDE EFFECT:
 */

char FConfig::get_key_prev_query ()
{
    return m_key_prev_query;
}

/*
 * FUNCTION:    FConfig::set_key_switch_mode
 *
 * DESCRIPTION: This function defines the key for switching between modes
 *
 * IN:          key             The key to map
 * IN-OUT:
 * OUT:
 * RETURN CODE: 0 if the key could be mapped.
 * SIDE EFFECT:
 */

int FConfig::set_key_switch_mode (char key)
{
    // Let's check if the key causes conflict.
    if (m_instance->is_valid_key(key, KEY_SWITCH_MODE) == false)
        return -1;
    
    m_instance->m_key_switch_mode = key;

    return 0;
}

/*
 * FUNCTION:    FConfig::get_key_switch_mode
 *
 * DESCRIPTION: This function returns the character corresponding to 
 *              the switch mode key.
 *
 * IN: 
 * IN-OUT:
 * OUT:
 * RETURN CODE: The switch mode key
 * SIDE EFFECT:
 */

char FConfig::get_key_switch_mode ()
{
    return m_key_switch_mode;
}

/*
 * FUNCTION:    FConfig::set_key_next_completion
 *
 * DESCRIPTION: This function defines the key for finding next symbol 
 *              completion.
 *
 * IN:          key             The key to map
 * IN-OUT:
 * OUT:
 * RETURN CODE: 0 if the key could be mapped.
 * SIDE EFFECT:
 */

int FConfig::set_key_next_completion (char key)
{
    // Let's check if the key causes conflict.
    if (m_instance->is_valid_key(key, KEY_NEXT_COMPLETION) == false)
        return -1;
    
    m_instance->m_key_next_completion = key;

    return 0;
}

/*
 * FUNCTION:    FConfig::get_key_next_completion
 *
 * DESCRIPTION: This function returns the character corresponding to 
 *              the next completion key.
 *
 * IN: 
 * IN-OUT:
 * OUT:
 * RETURN CODE: The next completion key
 * SIDE EFFECT:
 */

char FConfig::get_key_next_completion ()
{
    return m_key_next_completion;
}

/*
 * FUNCTION:    FConfig::set_key_prev_completion
 *
 * DESCRIPTION: This function defines the key for finding prev symbol 
 *              completion.
 *
 * IN:          key             The key to map
 * IN-OUT:
 * OUT:
 * RETURN CODE: 0 if the key could be mapped.
 * SIDE EFFECT:
 */

int FConfig::set_key_prev_completion (char key)
{
    // Let's check if the key causes conflict.
    if (m_instance->is_valid_key(key, KEY_PREV_COMPLETION) == false)
        return -1;
    
    m_instance->m_key_prev_completion = key;

    return 0;
}

/*
 * FUNCTION:    FConfig::get_key_prev_completion
 *
 * DESCRIPTION: This function returns the character corresponding to 
 *              the prev completion key.
 *
 * IN: 
 * IN-OUT:
 * OUT:
 * RETURN CODE: The prev completion key
 * SIDE EFFECT:
 */

char FConfig::get_key_prev_completion ()
{
    return m_key_prev_completion;
}

/*
 * FUNCTION:    FConfig::set_key_rebuild_db
 *
 * DESCRIPTION: This function defines the key for updating database
 *
 * IN:          key             The key to map
 * IN-OUT:
 * OUT:
 * RETURN CODE: 0 if the key could be mapped.
 * SIDE EFFECT:
 */

int FConfig::set_key_rebuild_db (char key)
{
    // Let's check if the key causes conflict.
    if (m_instance->is_valid_key(key, KEY_REBUILD_DB) == false)
        return -1;
    
    m_instance->m_key_rebuild_db = key;

    return 0;
}

/*
 * FUNCTION:    FConfig::get_key_rebuild_db
 *
 * DESCRIPTION: This function returns the character corresponding to 
 *              the rebuild database key.
 *
 * IN: 
 * IN-OUT:
 * OUT:
 * RETURN CODE: The rebuild database key
 * SIDE EFFECT:
 */

char FConfig::get_key_rebuild_db ()
{
    return m_key_rebuild_db;
}

/*
 * FUNCTION:    FConfig::set_key_help
 *
 * DESCRIPTION: This function defines the help key
 *
 * IN:          key             The key to map
 * IN-OUT:
 * OUT:
 * RETURN CODE: 0 if the key could be mapped.
 * SIDE EFFECT:
 */

int FConfig::set_key_help (char key)
{
    // Let's check if the key causes conflict.
    if (m_instance->is_valid_key(key, KEY_SHOW_HELP) == false)
        return -1;
    
    m_instance->m_key_help = key;

    return 0;
}

/*
 * FUNCTION:    FConfig::get_key_help
 *
 * DESCRIPTION: This function returns the character corresponding to 
 *              the help key.
 *
 * IN: 
 * IN-OUT:
 * OUT:
 * RETURN CODE: The help key
 * SIDE EFFECT:
 */

char FConfig::get_key_help ()
{
    return m_key_help;
}

/*
 * FUNCTION:    FConfig::set_key_escape
 *
 * DESCRIPTION: This function defines the escape key
 *
 * IN:          key             The key to map
 * IN-OUT:
 * OUT:
 * RETURN CODE: 0 if the key could be mapped.
 * SIDE EFFECT:
 */

int FConfig::set_key_escape (char key)
{
    // Let's check if the key causes conflict.
    if (m_instance->is_valid_key(key, KEY_ESCAPE) == false)
        return -1;
    
    m_instance->m_key_escape = key;

    return 0;
}

/*
 * FUNCTION:    FConfig::get_key_escape
 *
 * DESCRIPTION: This function returns the character corresponding to 
 *              the escape key.
 *
 * IN: 
 * IN-OUT:
 * OUT:
 * RETURN CODE: The escape key
 * SIDE EFFECT:
 */

char FConfig::get_key_escape ()
{
    return m_key_escape;
}

/*
 * FUNCTION:    FConfig::set_key_history_forward
 *
 * DESCRIPTION: This function defines the key for going forward into
 *              query history
 *
 * IN:          key             The key to map
 * IN-OUT:
 * OUT:
 * RETURN CODE: 0 if the key could be mapped.
 * SIDE EFFECT:
 */

int FConfig::set_key_history_forward (char key)
{
    // Let's check if the key causes conflict.
    if (m_instance->is_valid_key(key, KEY_HISTORY_FORWARD) == false)
        return -1;
    
    m_instance->m_key_history_forward = key;

    return 0;
}

/*
 * FUNCTION:    FConfig::get_key_history_forward
 *
 * DESCRIPTION: This function returns the character corresponding to 
 *              the forward history key.
 *
 * IN: 
 * IN-OUT:
 * OUT:
 * RETURN CODE: The escape key
 * SIDE EFFECT:
 */

char FConfig::get_key_history_forward ()
{
    return m_key_history_forward;
}

/*
 * FUNCTION:    FConfig::set_key_history_backward
 *
 * DESCRIPTION: This function defines the key for going backward into
 *              query history
 *
 * IN:          key             The key to map
 * IN-OUT:
 * OUT:
 * RETURN CODE: 0 if the key could be mapped.
 * SIDE EFFECT:
 */

int FConfig::set_key_history_backward (char key)
{
    // Let's check if the key causes conflict.
    if (m_instance->is_valid_key(key, KEY_HISTORY_BACKWARD) == false)
        return -1;
    
    m_instance->m_key_history_backward = key;

    return 0;
}

/*
 * FUNCTION:    FConfig::get_key_history_backward
 *
 * DESCRIPTION: This function returns the character corresponding to 
 *              the backward history key.
 *
 * IN: 
 * IN-OUT:
 * OUT:
 * RETURN CODE: The escape key
 * SIDE EFFECT:
 */

char FConfig::get_key_history_backward ()
{
    return m_key_history_backward;
}

/*
 * FUNCTION:    FConfig::set_dumping_db
 *
 * DESCRIPTION: This function is called if the program is invoked for
 *              dumping an existing database.
 *
 * IN:
 * IN-OUT:
 * OUT:
 * RETURN CODE:
 * SIDE EFFECT:
 */

void FConfig::set_dumping_db ()
{
    m_dumping_db = true;
}

/*
 * FUNCTION:    FConfig::get_dumping_db
 *
 * DESCRIPTION: This function returns true if the program is invoked for
 *              dumping a database.
 *
 * IN:
 * IN-OUT:
 * OUT:
 * RETURN CODE: true if dumping existing database
 * SIDE EFFECT:
 */

bool FConfig::get_dumping_db ()
{
    return m_dumping_db;
}
